import logging
from argparse import ArgumentParser

from monobiome.cli import fill, scheme, palette

logger: logging.Logger = logging.getLogger(__name__)

def configure_logging(log_level: int) -> None:
    """
    Configure logger's logging level.
    """

    logger.setLevel(log_level)

def create_parser() -> ArgumentParser:
    parser = ArgumentParser(
        description="Accent modeling CLI",
    )
    parser.add_argument(
        "--log-level",
        type=int,
        metavar="int",
        choices=[10, 20, 30, 40, 50],
        help="Log level: 10=DEBUG, 20=INFO, 30=WARNING, 40=ERROR, 50=CRITICAL",
    )

    subparsers = parser.add_subparsers(help="subcommand help")

    fill.register_parser(subparsers)
    scheme.register_parser(subparsers)
    palette.register_parser(subparsers)

    return parser



from pathlib import Path
from argparse import Namespace, ArgumentParser

from monobiome.util import _SubparserType
from monobiome.palette import generate_palette


def register_parser(subparsers: _SubparserType) -> None:
    parser = subparsers.add_parser(
        "palette",
        help="generate primary palette"
    )

    parser.add_argument(
        "-n",
        "--notation",
        type=str,
        default="hex",
        choices=["hex", "oklch"],
        help="color notation to export (either hex or oklch)",
    )
    parser.add_argument(
        "-f",
        "--format",
        type=str,
        default="toml",
        choices=["json", "toml"],
        help="format of palette file (either JSON or TOML)",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="output file to write palette content",
    )

    parser.set_defaults(func=handle_palette)


def handle_palette(args: Namespace, parser: ArgumentParser) -> None:
    notation = args.notation
    file_format = args.format
    output = args.output

    palette_text = generate_palette(notation, file_format)

    if output is None:
        print(palette_text)
    else:
        with Path(output).open("w") as f:
            f.write(palette_text)



from pathlib import Path
from argparse import Namespace, ArgumentParser

from monobiome.util import _SubparserType
from monobiome.scheme import generate_scheme
from monobiome.constants import monotone_h_map


def register_parser(subparsers: _SubparserType) -> None:
    parser = subparsers.add_parser(
        "scheme",
        help="create scheme variants"
    )

    parser.add_argument(
        "mode",
        type=str,
        choices=["dark", "light"],
        help="scheme mode (light or dark)"
    )
    parser.add_argument(
        "biome",
        type=str,
        choices=list(monotone_h_map.keys()),
        help="biome setting for scheme"
    )
    parser.add_argument(
        "-m",
        "--metric",
        type=str,
        default="oklch",
        choices=["wcag", "oklch", "lightness"],
        help="metric to use for measuring swatch distances"
    )

    # e.g., wcag=4.5; oklch=0.40; lightness=40
    parser.add_argument(
        "-d",
        "--distance",
        type=float,
        default=0.40,
        help="distance threshold for specified metric",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="output file to write scheme content",
    )

    # these params remain rooted in lightness; no need to accommodate metric
    # given these are monotone adjustments. You *could* consider rooting these
    # in metric units, but along monotones, distance=lightness and WCAG isn't a
    # particularly good measure of perceptual distinction, so we'd prefer the
    # former.
    parser.add_argument(
        "-l",
        "--l-base",
        type=int,
        default=20,
        help="minimum lightness level (default: 20)",
    )
    parser.add_argument(
        "--l-step",
        type=int,
        default=5,
        help="lightness step size (default: 5)",
    )

    # gaps
    parser.add_argument(
        "--fg-gap",
        type=int,
        default=50,
        help="foreground lightness gap (default: 50)",
    )
    parser.add_argument(
        "--grey-gap",
        type=int,
        default=30,
        help="grey lightness gap (default: 30)",
    )
    parser.add_argument(
        "--term-fg-gap",
        type=int,
        default=65,
        help="terminal foreground lightness gap (default: 60)",
    )

    parser.set_defaults(func=handle_scheme)


def handle_scheme(args: Namespace, parser: ArgumentParser) -> None:
    output = args.output

    mode = args.mode
    biome =  args.biome
    metric = args.metric
    distance = args.distance
    l_base = args.l_base
    l_step = args.l_step
    fg_gap = args.fg_gap
    grey_gap = args.grey_gap
    term_fg_gap = args.term_fg_gap

    full_color_map = {
        "red": "red",
        "orange": "orange",
        "yellow": "yellow",
        "green": "green",
        "cyan": "cyan",
        "blue": "blue",
        "violet": "violet",
        "magenta": "orange",
    }
    term_color_map = {
        "red": "red",
        "yellow": "yellow",
        "green": "green",
        "cyan": "blue",
        "blue": "blue",
        "magenta": "orange",
    }
    vim_color_map = {
        "red": "red",
        "orange": "orange",
        "yellow": "yellow",
        "green": "green",
        "cyan": "green",
        "blue": "blue",
        "violet": "blue",
        "magenta": "red",
    }
    # vim_color_map = full_color_map

    scheme_text = generate_scheme(
        mode,
        biome,
        metric,
        distance,
        l_base,
        l_step,
        fg_gap,
        grey_gap,
        term_fg_gap,
        full_color_map,
        term_color_map,
        vim_color_map,
    )

    if output is None:
        print(scheme_text)
    else:
        with Path(output).open("w") as f:
            f.write(scheme_text)



import tomllib
from importlib.resources import files

import numpy as np

from monobiome.curve import (
    l_maxC_h,
    bezier_y_at_x,
)

parameters_file = files("monobiome.data") / "parameters.toml"
parameters = tomllib.load(parameters_file.open("rb"))

L_min: int = parameters.get("L_min", 10)
L_max: int = parameters.get("L_max", 98)
L_step: int = parameters.get("L_step", 5)

L_points: list[int] = list(range(L_min, L_max+1))

# L-space just affects accuracy of chroma max
L_space = np.arange(0, 100 + L_step, L_step)

monotone_C_map = parameters.get("monotone_C_map", {})
h_weights = parameters.get("h_weights", {})
h_L_offsets = parameters.get("h_L_offsets", {})
h_C_offsets = parameters.get("h_C_offsets", {})
monotone_h_map = parameters.get("monotone_h_map", {})
accent_h_map = parameters.get("accent_h_map", {})
h_map = {**monotone_h_map, **accent_h_map}

"""
Compute chroma maxima at provided lightness levels across hues.

A map with max chroma values for each hue across lightness space

{
   "red": [ Cmax@L=10, Cmax@L=11, Cmax@L=12, ... ],
   "orange": [ Cmax@L=10, Cmax@L=11, Cmax@L=12, ... ],
   ...
}
"""
Lspace_Cmax_Hmap = {
    h_str: [l_maxC_h(_L, _h) for _L in L_space]
    for h_str, _h in h_map.items()
}


"""
Set QBR curves, *unbounded* chroma curves for all hues

1. Raw bezier chroma values for each hue across the lightness space

   Lpoints_Cqbr_Hmap = {
      "red": [ Bezier@L=10, Bezier@L=11, Bezier@L=12, ... ],
      ...
   }

2. Three bezier control points for each hue's chroma curve

   QBR_ctrl_Hmap = {
      "red": np.array([
          [ x1, y1 ],
          [ x2, y2 ],
          [ x3, y3 ]
       ]),
      ...
   }
"""
Lpoints_Cqbr_Hmap = {}
QBR_ctrl_Hmap = {}

for h_str, _h in monotone_h_map.items():
    Lpoints_Cqbr_Hmap[h_str] = np.array(
        [monotone_C_map[h_str]]*len(L_points)
    )
    
for h_str, _h in accent_h_map.items():
    Lspace_Cmax = Lspace_Cmax_Hmap[h_str]
    
    # get L value of max chroma; will be a bezier control
    L_Cmax_idx = np.argmax(Lspace_Cmax)
    L_Cmax = L_space[L_Cmax_idx]

    # offset control point by any preset x-shift
    L_Cmax += h_L_offsets[h_str]

    # and get max C at the L offset
    Cmax = l_maxC_h(L_Cmax, _h)

    # set 3 control points; shift by any global linear offest
    C_offset = h_C_offsets.get(h_str, 0)
    
    p_0 = np.array([0, 0])
    p_Cmax = np.array([L_Cmax, Cmax + C_offset])
    p_100 = np.array([100, 0])
    
    B_L_points = bezier_y_at_x(
        p_0, p_Cmax, p_100,
        h_weights.get(h_str, 1),
        L_points
    )
    Lpoints_Cqbr_Hmap[h_str] = B_L_points
    QBR_ctrl_Hmap[h_str] = np.vstack([p_0, p_Cmax, p_100])


"""
Bezier chroma values, but bounded to attainable gamut colors (bezier fit
can produce invalid chroma values)

h_L_points_Cstar = {
   "red": [ bounded-bezier@L=10, bounded-bezier@L=11, ... ],
   ...
}
"""
Lpoints_Cstar_Hmap = {}

for h_str, L_points_C in Lpoints_Cqbr_Hmap.items():
    _h = h_map[h_str]

    Lpoints_Cstar_Hmap[h_str] = [
        max(0, min(_C, l_maxC_h(_L, _h)))
        for _L, _C in zip(L_points, L_points_C, strict=True)
    ]


# strictly enforce curve bounds s.t. there are no intersections
# order is determined by the max attained chromap
max_Cstar_Horder = [
    (h_str, max(Lpoints_Cstar))
    for h_str, Lpoints_Cstar in Lpoints_Cstar_Hmap.items()
]
max_Cstar_Horder = sorted(max_Cstar_Horder, key=lambda t: t[1], reverse=True)

for i in range(len(max_Cstar_Horder)-1):
    outer_h, _ = max_Cstar_Horder[i]
    inner_h, _ = max_Cstar_Horder[i+1]

    Lpoints_Cstar_Hmap[inner_h] = [
        min(inner_c, Lpoints_Cstar_Hmap[outer_h][ci])
        for ci, inner_c in enumerate(Lpoints_Cstar_Hmap[inner_h])
    ]



import math
from types import GenericAlias
from argparse import ArgumentParser, _SubParsersAction

from coloraide import Color

_SubParsersAction.__class_getitem__ = classmethod(GenericAlias)
_SubparserType = _SubParsersAction[ArgumentParser]

def oklch_distance(xc: Color, yc: Color) -> float:
    """
    Compute the distance between two colors in OKLCH space.

    Note: `xc` and `yc` are presumed to be OKLCH colors already, such that
    `.coords()` yields an `(l, c, h)` triple directly rather than first
    requiring conversion. When we can make this assumption, we save roughly an
    order of magnitude in runtime.

    1. `xc.distance(yc, space="oklch")`: 500k evals takes ~2s
    2. This method: 500k evals takes ~0.2s
    """

    l1, c1, h1 = xc.coords()
    l2, c2, h2 = yc.coords()

    rad1 = h1 / 180 * math.pi
    rad2 = h2 / 180 * math.pi
    x1, y1 = c1 * math.cos(rad1), c1 * math.sin(rad1)
    x2, y2 = c2 * math.cos(rad2), c2 * math.sin(rad2)
    
    dx = x1 - x2
    dy = y1 - y2
    dz = l1 - l2
    
    return (dx**2 + dy**2 + dz**2)**0.5



import sys
import tomllib
from pathlib import Path
from argparse import Namespace, ArgumentParser

from symconf.template import Template, TOMLTemplate

from monobiome.util import _SubparserType
from monobiome.palette import generate_palette


def register_parser(subparsers: _SubparserType) -> None:
    parser = subparsers.add_parser(
        "fill",
        help="fill scheme templates for applications"
    )

    parser.add_argument(
        "scheme",
        type=str,
        help="scheme file path"
    )
    parser.add_argument(
        "template",
        nargs="?",
        help="template file path (defaults to stdin)"
    )
    parser.add_argument(
        "-p",
        "--palette",
        type=str,
        help="palette file to use for color definitions",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="output file to write filled template",
    )

    parser.set_defaults(func=handle_fill)


def handle_fill(args: Namespace, parser: ArgumentParser) -> None:
    scheme = Path(args.scheme)
    output = args.output

    if args.template is None:
        if sys.stdin.isatty():
            parser.error("no template provided (file or stdin)")
            return
        template_content = sys.stdin.read()
    elif args.template == "-":
        template_content = sys.stdin.read()
    else:
        try:
            template_content = Path(args.template).read_text()
        except OSError as e:
            parser.error(f"cannot read template file: {e}")

    if args.palette:
        try:
            palette_toml = Path(args.palette).read_text()
        except OSError as e:
            parser.error(f"cannot read palette file: {e}")
            return
    else:
        palette_toml = generate_palette("hex", "toml")

    palette_dict = tomllib.loads(palette_toml)
    concrete_scheme = TOMLTemplate(scheme).fill_dict(palette_dict)
    filled_template = Template(template_content).fill(concrete_scheme)

    if output is None:
        print(filled_template)
    else:
        with Path(output).open("w") as f:
            f.write(filled_template)



from monobiome.cli import create_parser, configure_logging


def main() -> None:
    parser = create_parser()
    args = parser.parse_args()

    # skim off log level to handle higher-level option
    if hasattr(args, "log_level") and args.log_level is not None:
        configure_logging(args.log_level)

    if "func" in args:
        args.func(args, parser)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



import json
from typing import Any
from functools import cache
from importlib.metadata import version

from coloraide import Color

from monobiome.constants import (
    h_map,
    L_points,
    Lpoints_Cstar_Hmap,
)


@cache
def compute_hlc_map(notation: str) -> dict[str, Any]:
    hlc_map = {}

    for h_str, Lpoints_Cstar in Lpoints_Cstar_Hmap.items():
        _h = h_map[h_str]
        hlc_map[h_str] = {}
        
        for _l, _c in zip(L_points, Lpoints_Cstar, strict=True):
            oklch = Color('oklch', [_l/100, _c, _h])

            if notation == "hex":
                srgb = oklch.convert('srgb')
                c_str = srgb.to_string(hex=True)
            elif notation == "oklch":
                ol, oc, oh = oklch.convert('oklch').coords()
                c_str = f"oklch({ol*100:.1f}% {oc:.4f} {oh:.1f})"
            
            hlc_map[h_str][_l] = c_str

    return hlc_map

def generate_palette(
    notation: str,
    file_format: str,
) -> str:
    mb_version = version("monobiome")
    hlc_map = compute_hlc_map(notation)
            
    if file_format == "json":
        hlc_map["version"] = mb_version
        return json.dumps(hlc_map, indent=4)
    else:
        toml_lines = [f"version = \"{mb_version}\"", ""]
        for _h, _lc_map in hlc_map.items():
            toml_lines.append(f"[{_h}]")
            for _l, _c in _lc_map.items():
                toml_lines.append(f'l{_l} = "{_c}"')
            toml_lines.append("")

        return "\n".join(toml_lines)



from functools import cache
from collections.abc import Callable
from importlib.metadata import version

from coloraide import Color

from monobiome.util import oklch_distance
from monobiome.palette import compute_hlc_map
from monobiome.constants import (
    accent_h_map,
    monotone_h_map,
)


@cache
def compute_dma_map(
    dT: float,
    metric: Callable | None = None
) -> dict[str, dict]:
    """
    For threshold `dT`, compute the nearest accent shades that exceed that
    threshold for every monotone shade.

    Returns: map of minimum constraint satisfying accent colors for monotone
        spectra

        {
            "alpine": {
                "oklch( ... )": {
                    "red": *nearest oklch >= dT from M base*,
                    ...
                },
                ...
            },
            ...
        }
    """

    if metric is None:
        metric = oklch_distance
    
    oklch_hlc_map = compute_hlc_map("oklch")
    oklch_color_map = {
        c_name: [Color(c_str) for c_str in c_str_dict.values()]
        for c_name, c_str_dict in oklch_hlc_map.items()
    }
    
    dT_mL_acol_map = {}
    for m_name in monotone_h_map:
        mL_acol_map = {}
        m_colors = oklch_color_map[m_name]
        
        for m_color in m_colors:
            acol_min_map = {}
            
            for a_name in accent_h_map:
                a_colors = oklch_color_map[a_name]
                oklch_dists = filter(
                    lambda d: (d[1] - dT) >= 0,
                    [
                        (ac, metric(m_color, ac))
                        for ac in a_colors
                    ]
                )
                oklch_dists = list(oklch_dists)
                if oklch_dists:
                    min_a_color = min(oklch_dists, key=lambda t: t[1])[0]
                    acol_min_map[a_name] = min_a_color
    
            # make sure the current monotone level has *all* accents; o/w
            # ignore
            if len(acol_min_map) < len(accent_h_map):
                continue

            mL = m_color.coords()[0]
            mL_acol_map[int(mL*100)] = acol_min_map
        dT_mL_acol_map[m_name] = mL_acol_map

    return dT_mL_acol_map

def generate_scheme_groups(
    mode: str,
    biome: str,
    metric: str,
    distance: float,
    l_base: int,
    l_step: int,
    fg_gap: int,
    grey_gap: int,
    term_fg_gap: int,
    accent_color_map: dict[str, str],
) -> tuple[list[tuple[str, str]], ...]:
    """
    Parameters:
        mode: one of ["dark", "light"]
        biome: biome setting
        metric: one of ["wcag", "oklch", "lightness"]
    """

    metric_map = {
        "wcag": lambda mc,ac: ac.contrast(mc, method='wcag21'),
        "oklch": oklch_distance,
        "lightness": lambda mc,ac: abs(mc.coords()[0]-ac.coords()[0])*100,
    }
    
    metric_func = metric_map[metric]
    dT_mL_acol_map = compute_dma_map(distance, metric=metric_func)
    Lma_map = {
        m_name: mL_acol_dict[l_base]
        for m_name, mL_acol_dict in dT_mL_acol_map.items()
        if l_base in mL_acol_dict
    }
    
    # the `mL_acol_dict` only includes lightnesses where all accent colors were
    # within threshold. Coverage here will be partial if, at the `mL`, there is
    # some monotone base that doesn't have all accents within threshold. This
    # can happen at the edge, e.g., alpine@L15 has all accents w/in the
    # distance, but the red accent was too far under tundra@L15, so there's no
    # entry. This particular case is fairly rare; it's more likely that *all*
    # monotones are undefined. Either way, both such cases lead to partial
    # scheme coverage.
    if len(Lma_map) < len(monotone_h_map):
        print(f"Warning: partial scheme coverage for {l_base=}@{distance=}")
    if biome not in Lma_map:
        print(f"Biome {biome} unable to meet {metric} constraints")
    accent_colors = Lma_map.get(biome, {})
    
    meta_pairs = [
        ("version", version("monobiome")),
        ("mode", mode),
        ("biome", biome),
        ("metric", metric),
        ("distance", str(distance)),
        ("l_base", str(l_base)),
        ("l_step", str(l_step)),
        ("fg_gap", str(fg_gap)),
        ("grey_gap", str(grey_gap)),
        ("term_fg_gap", str(term_fg_gap)),
    ]

    # note how selection_bg steps up by `l_step`, selection_fg steps down by
    # `l_step` (from their respective bases)
    term_pairs = [
        ("background", f"f{{{{{biome}.l{l_base}}}}}"),
        ("selection_bg", f"f{{{{{biome}.l{l_base+l_step}}}}}"),
        ("selection_fg", f"f{{{{{biome}.l{l_base+term_fg_gap-l_step}}}}}"),
        ("foreground", f"f{{{{{biome}.l{l_base+term_fg_gap}}}}}"),
        ("cursor", f"f{{{{{biome}.l{l_base+term_fg_gap-l_step}}}}}"),
        ("cursor_text", f"f{{{{{biome}.l{l_base+l_step}}}}}"),
    ]

    monotone_pairs = []
    monotone_pairs += [
        (f"bg{i}", f"f{{{{{biome}.l{l_base+i*l_step}}}}}")
        for i in range(4)
    ]
    monotone_pairs += [
        (f"fg{3-i}", f"f{{{{{biome}.l{fg_gap+l_base+i*l_step}}}}}")
        for i in range(4)
    ]
 
    accent_pairs = [
        ("black", f"f{{{{{biome}.l{l_base}}}}}"),
        ("grey", f"f{{{{{biome}.l{l_base+grey_gap}}}}}"),
        ("white", f"f{{{{{biome}.l{l_base+term_fg_gap-2*l_step}}}}}"),
    ]
    for color_name, mb_accent in accent_color_map.items():
        aL = int(100*accent_colors[mb_accent].coords()[0])
        accent_pairs.append(
            (
                color_name,
                f"f{{{{{mb_accent}.l{aL}}}}}"
            )
        )
    
    return meta_pairs, term_pairs, monotone_pairs, accent_pairs

def generate_scheme(
    mode: str,
    biome: str,
    metric: str,
    distance: float,
    l_base: int,
    l_step: int,
    fg_gap: int,
    grey_gap: int,
    term_fg_gap: int,
    full_color_map: dict[str, str],
    term_color_map: dict[str, str],
    vim_color_map: dict[str, str],
) -> str:
    l_sys = l_base
    l_app = l_base + l_step

    term_bright_offset = 10

    # negate gaps if mode is light
    if mode == "light":
        l_step *= -1
        fg_gap *= -1
        grey_gap *= -1
        term_fg_gap *= -1
        term_bright_offset *= -1

    meta, _, mt, ac = generate_scheme_groups(
        mode, biome, metric, distance,
        l_sys, l_step,
        fg_gap, grey_gap, term_fg_gap,
        full_color_map
    )

    _, term, _, term_norm_ac = generate_scheme_groups(
        mode, biome, metric, distance,
        l_app, l_step,
        fg_gap, grey_gap, term_fg_gap,
        term_color_map
    )
    _, _, _, term_bright_ac = generate_scheme_groups(
        mode, biome, metric, distance,
        l_app + term_bright_offset, l_step,
        fg_gap, grey_gap, term_fg_gap,
        term_color_map
    )

    _, _, vim_mt, vim_ac = generate_scheme_groups(
        mode, biome, metric, distance,
        l_app, l_step,
        fg_gap, grey_gap, term_fg_gap,
        vim_color_map
    )

    def pair_strings(pair_list: list[tuple[str, str]]) -> list[str]:
        return [
            f"{lhs:<12} = \"{rhs}\""
            for lhs, rhs in pair_list
        ]

    mb_version = version("monobiome")
    scheme_pairs = [
        "# ++ monobiome scheme file ++",
        f"# ++ generated CLI @ {mb_version} ++",
    ]
    scheme_pairs += pair_strings(meta)
    scheme_pairs += pair_strings(mt)
    scheme_pairs += pair_strings(ac)

    scheme_pairs += ["", "[term]"]
    scheme_pairs += pair_strings(term)

    scheme_pairs += ["", "[term.normal]"]
    scheme_pairs += pair_strings(term_norm_ac)

    scheme_pairs += ["", "[term.bright]"]
    scheme_pairs += pair_strings(term_bright_ac)

    scheme_pairs += ["", "[vim]"]
    scheme_pairs += pair_strings(vim_mt)
    scheme_pairs += pair_strings(vim_ac)

    return "\n".join(scheme_pairs)



from importlib.metadata import version

import numpy as np
import matplotlib.pyplot as plt
from coloraide import Color
from matplotlib.collections import LineCollection

from monobiome.palette import compute_hlc_map
from monobiome.constants import (
    h_map,
    L_space,
    L_points,
    accent_h_map,
    monotone_h_map,
    Lspace_Cmax_Hmap,
    max_Cstar_Horder,
    Lpoints_Cstar_Hmap,
)

VERSION = version("monobiome")

def plot_hue_chroma_bounds() -> tuple[plt.Figure, plt.Axes]:
    name_h_map = {}
    ax_h_map = {}
    fig, axes = plt.subplots(
        len(monotone_h_map),
        1,
        sharex=True,
        sharey=True,
        figsize=(4, 10)
    )

    for i, h_str in enumerate(Lpoints_Cstar_Hmap):
        _h = h_map[h_str]

        l_space_Cmax = Lspace_Cmax_Hmap[h_str]
        l_points_Cstar = Lpoints_Cstar_Hmap[h_str]
        
        if _h not in ax_h_map:
            ax_h_map[_h] = axes[i]
        ax = ax_h_map[_h]
        
        if _h not in name_h_map:
            name_h_map[_h] = []
        name_h_map[_h].append(h_str)

        # plot Cmax and Cstar
        ax.plot(L_space, l_space_Cmax, c="g", alpha=0.3, label="Cmax")
        
        cstar_label = f"{'accent' if h_str in accent_h_map else 'monotone'} C*"
        ax.plot(L_points, l_points_Cstar, alpha=0.7, label=cstar_label)
        
        ax.title.set_text(f"Hue [${_h}$] - {'|'.join(name_h_map[_h])}")
        
    axes[-1].set_xlabel("Lightness (%)")
    axes[-1].set_xticks([L_points[0], L_points[-1]])

    fig.tight_layout()
    fig.subplots_adjust(top=0.9)

    handles, labels = axes[-1].get_legend_handles_labels()
    unique = dict(zip(labels, handles, strict=True))
    fig.legend(
        unique.values(),
        unique.keys(),
        loc='lower center',
        bbox_to_anchor=(0.5, -0.06),
        ncol=3
    )

    plt.suptitle(f"$C^*$ curves for hue groups (v{VERSION})")

    return fig, axes

def plot_hue_chroma_star() -> tuple[plt.Figure, plt.Axes]:
    fig, ax = plt.subplots(1, 1, figsize=(8, 6))

    colors = accent_h_map.keys()
    # uncomment to preview just the 5 core term colors
    # colors = set(["red", "orange", "yellow", "green", "blue"])

    for h_str, _ in max_Cstar_Horder:
        Lpoints_Cstar = Lpoints_Cstar_Hmap[h_str]

        if h_str not in accent_h_map or h_str not in colors:
            continue

        _h = h_map[h_str]
        h_colors = [
            Color('oklch', [_l/100, _c, _h]).convert("srgb")
            for _l, _c in zip(L_points, Lpoints_Cstar, strict=True)
        ]
        
        x = np.asarray(L_points)
        y = np.asarray(Lpoints_Cstar)
        pts = np.column_stack([x, y]).reshape(-1, 1, 2)
        segs = np.concatenate([pts[:-1], pts[1:]], axis=1)
        rgb = np.asarray(h_colors)
        seg_colors = (rgb[:-1] + rgb[1:]) / 2
        lc = LineCollection(segs, colors=seg_colors, linewidth=3,
                            capstyle="round", joinstyle="round",
                            label=h_str)

        ax.add_collection(lc)
        ax.autoscale_view()
        
    ax.set_xlabel("Lightness (%)")
    ax.set_xticks([L_points[0], 45, 50, 55, 60, 65, 70, L_points[-1]])

    plt.suptitle(f"$C^*$ curves (v{VERSION})")

    return fig, ax

def palette_image(
    palette: dict[str, dict[int, str]],
    cell_size: int = 40,
    keys: list[str] | None = None
) -> tuple[np.ndarray, list[str], list[list[int]]]:
    names = list(palette.keys()) if keys is None else keys
        
    row_count = len(names)
    col_counts = [len(palette[n]) for n in names]
    max_cols = max(col_counts)

    h = row_count * cell_size
    w = max_cols * cell_size
    img = np.ones((h, w, 3), float)

    lightness_keys_per_row = []

    for r, name in enumerate(names):
        shades = palette[name]
        lkeys = sorted(shades.keys())
        lightness_keys_per_row.append(lkeys)
        for c, k in enumerate(lkeys):
            col = Color(shades[k]).convert("srgb").fit(method="clip")
            rgb = [col["r"], col["g"], col["b"]]
            r0, r1 = r * cell_size, (r + 1) * cell_size
            c0, c1 = c * cell_size, (c + 1) * cell_size
            img[r0:r1, c0:c1, :] = rgb

    return img, names, lightness_keys_per_row

def show_palette(
    palette: dict[str, dict[int, str]],
    cell_size: int = 40,
    keys: list[str] | None = None,
    show_labels: bool = True,
    dpi: int = 100,
) -> tuple[plt.Figure, plt.Axes]:
    img, names, _ = palette_image(palette, cell_size, keys=keys)

    fig_w = img.shape[1] / 100
    fig_h = img.shape[0] / 100

    if show_labels:
        fig, ax = plt.subplots(figsize=(fig_w, fig_h), dpi=dpi)

        ax.imshow(img, interpolation="none", origin="upper")
        ax.set_xticks([])

        ytick_pos = [(i + 0.5) * cell_size for i in range(len(names))]
        ax.set_yticks(ytick_pos)
        ax.set_yticklabels(names)
        ax.set_ylim(img.shape[0], 0)  # ensures rows render w/o half-cells

        return fig, ax

    fig = plt.figure(figsize=(fig_w, fig_h), dpi=dpi, frameon=False)
    ax = fig.add_axes((0, 0, 1, 1), frame_on=False)
    ax.imshow(
        img,
        interpolation="nearest",
        origin="upper",
        extent=(0, img.shape[1], img.shape[0], 0),
        aspect="auto",
    )
    ax.set_xlim(0, img.shape[1])
    ax.set_ylim(img.shape[0], 0)
    ax.axis("off")
    fig.subplots_adjust(0, 0, 1, 1, hspace=0, wspace=0)

    return fig, ax

if __name__ == "__main__":
    keys = [
        "alpine",
        "badlands",
        "chaparral",
        "savanna",
        "grassland",
        "reef",
        "tundra",
        "heathland",
        "moorland",
        "orange",
        "yellow",
        "green",
        "cyan",
        "blue",
        "violet",
        "magenta",
        "red",
    ]
    term_keys = [
        "alpine",
        "badlands",
        "chaparral",
        "savanna",
        "grassland",
        "tundra",
        "red",
        "orange",
        "yellow",
        "green",
        "blue",
    ]

    hlc_map = compute_hlc_map("oklch")
    show_palette(hlc_map, cell_size=25, keys=keys)
    # show_palette(hlc_map, cell_size=1, keys=term_keys)



from functools import cache

import numpy as np
from coloraide import Color


def quad_bezier_rational(
    P0: float,
    P1: float,
    P2: float,
    w: float,
    t: np.ndarray,
) -> np.ndarray:
    """
    Compute the point values of a quadratic rational Bezier curve.

    Uses `P0`, `P1`, and `P2` as the three control points of the curve. `w`
    controls the weight toward the middle control point ("sharpness" of the
    curve"), and `t` is the number of sample points used along the curve.
    """

    t = np.asarray(t)[:, None]
    num = (1-t)**2*P0 + 2*w*(1-t)*t*P1 + t**2*P2
    den = (1-t)**2 + 2*w*(1-t)*t + t**2
    
    return num / den
    
def bezier_y_at_x(
    P0: float,
    P1: float,
    P2: float,
    w: float,
    x: float,
    n: int = 400,
) -> np.ndarray:
    """
    For the provided QBR parameters, provide the curve value at the given
    input.
    """

    t = np.linspace(0, 1, n)
    B = quad_bezier_rational(P0, P1, P2, w, t)
    x_vals, y_vals = B[:, 0], B[:, 1]
    
    return np.interp(x, x_vals, y_vals)

@cache
def l_maxC_h(
    _l: float,
    _h: float,
    space: str = 'srgb',
    eps: float = 1e-6,
    tol: float = 1e-9
) -> float:
    """
    Binary search for max attainable OKLCH chroma at fixed lightness and hue.

    Parameters:
        _l: lightness
        _h: hue

    Returns:
        Max in-gamut chroma at provided lightness and hue
    """

    def chroma_in_gamut(_c: float) -> bool:
        color = Color('oklch', [_l/100, _c, _h])
        return color.convert(space).in_gamut(tolerance=tol)

    lo, hi = 0.0, 0.1
    while chroma_in_gamut(hi):
        hi *= 2
    while hi - lo > eps:
        m = (lo + hi) / 2
        lo, hi = (m, hi) if chroma_in_gamut(m) else (lo, m)

    return lo
